# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import requests
from bag.check_rst import check_rst_file
from bag.console import bool_input
from nine import input, nine
from path import path as pathpy
from . import ReleaseStep, StopRelease, CommandStep
from .regex import version_in_python_source_file

__all__ = ('Shell', 'CheckRstFiles', 'InteractivelyApproveDistribution',
     'InteractivelyEnsureChangesDocumented', 'CheckTravis',
    'SetVersionNumberInteractively', 'PypiRegister', 'PypiUpload',
    'SetFutureVersion', 'ErrorStep')


@nine
class Shell(ReleaseStep):
    '''Runs some shell command.'''
    ERROR_CODE = 2

    def __init__(self, command, stop_on_failure=True):
        self.COMMAND = command
        self.stop_on_failure = stop_on_failure
        self.no_rollback = 'Unable to roll back the step {0}'.format(self)

    def __call__(self):
        self._execute_or_complain(self.COMMAND)  # sets self.success

    def __str__(self):
        return '[' + self.COMMAND + ']'


class CheckRstFiles(ReleaseStep):
    '''Helps keep documentation correct by verifying .rst documentation files.
    If files are not provided to the constructor, recursively finds *.rst files
    in the current directory and subdirectories.
    '''
    ERROR_CODE = 4

    def __init__(self, *files):
        self.paths = files

    def __call__(self):
        paths = self.paths or pathpy('.').walkfiles('*.rst')
        for path in paths:
            self.log.info('Checking ' + path)
            warnings = check_rst_file(path)
            if warnings:
                raise StopRelease('There are errors in {0}:\n{1}'.format(path,
                    '\n'.join([str(w) for w in warnings])))


class InteractivelyApproveDistribution(ReleaseStep):
    '''Generate a source distribution and let the user verify that
    all files are in there.
    '''
    ERROR_CODE = 5

    def __call__(self):
        self._execute_or_complain('python setup.py sdist')
        # TODO: Optionally xdg-open the archive for convenience
        # Create the sdist with "-d <output_dir>"  on a temp dir.
        # Delete it at the end.
        print("A temporary source distribution has been generated. This is")
        print("your chance to open the new archive (in the 'dist' directory)")
        print("and check that all files are in there.")
        if not bool_input("Do you approve the archive contents?"):
            raise StopRelease('Source distribution content not approved.\n'
                'If the distribution is missing some files,\n'
                "try correcting your MANIFEST.in file according to\n"
                "http://docs.python.org/3/distutils/sourcedist.html"
                "#specifying-the-files-to-distribute")


class InteractivelyEnsureChangesDocumented(ReleaseStep):
    ERROR_CODE = 3

    def __call__(self):
        if bool_input('Did you remember to update the CHANGES file?'):
            self.log.debug('User says CHANGES file is up to date.')
        else:
            raise StopRelease('One more joshlabotniked release is avoided.')


class CheckTravis(ReleaseStep):
    '''Checks the status, on travis-ci.org, of the latest build of a
    software project in a certain branch.
    '''
    ERROR_CODE = 91
    URL = 'https://api.travis-ci.org/repos/' \
          '{github_user}/{github_repository}/builds'

    def __call__(self):
        branch = self.config.get('branch', 'master')
        resp = requests.get(self.URL.format(**self.config))
        builds = resp.json()
        onbranch = filter(lambda x: x['branch'] == branch, builds)
        finished = filter(lambda x: x['state'] == 'finished', onbranch)
        if len(list(finished)) == 0:
            raise StopRelease(
                'Travis has not built branch "{0}" yet.'.format(branch))
        latest = finished[0]
        if latest['result'] == 0:
            self.log.info('No problem in latest Travis build: "{0}"'.format(
                          latest.get('message')))
            self._succeed()
        else:
            raise StopRelease(
                'Last Travis build on branch "{0}" failed.'.format(branch))


class SetVersionNumberInteractively(ReleaseStep):
    '''Asks the user for the new version number and writes it onto
    the configured source code file.
    '''
    ERROR_CODE = 6

    def __call__(self):
        releaser = self.releaser
        path = releaser.config['version_file']
        keyword = releaser.config['version_keyword']
        releaser.old_version = version_in_python_source_file(
            path, keyword=keyword)
        print('Current version: {0}'.format(releaser.old_version))
        releaser.the_version = input('What is the new version number? ')
        # Write the new version onto the source code
        version_in_python_source_file(
            path, keyword=keyword, replace=releaser.the_version)
        self._succeed()


class PypiRegister(CommandStep):
    COMMAND = 'python setup.py register'
    ERROR_CODE = 7
    no_rollback = 'The release has been created on http://pypi.python.org,\n' \
        'Unfortunately, the release must be removed manually.'

    def _validate_command_output(self, command_output):
        return "Server response (200): OK" in command_output


class PypiUpload(CommandStep):
    COMMAND = 'python setup.py sdist upload'
    ERROR_CODE = 8
    no_rollback = 'Cannot roll back the sdist upload to http://pypi.python.org'

    def _validate_command_output(self, command_output):
        return "Server response (200): OK" in command_output


class SetFutureVersion(ReleaseStep):
    '''Replaces the version number in the configured source code file with
    the future version number (the one ending in 'dev').
    '''
    ERROR_CODE = 9

    def __call__(self):
        releaser = self.releaser
        path = releaser.config['version_file']
        keyword = releaser.config['version_keyword']
        # If the SetVersionNumberInteractively step is disabled for debugging,
        # we can still execute the current step, by populating the_version:
        if releaser.the_version is None:
            releaser._the_version = version_in_python_source_file(
                path, keyword=keyword)
        self.log.info('Ready for the next development cycle! Setting version '
              + releaser.future_version)
        version_in_python_source_file(
            path, keyword=keyword, replace=releaser.future_version)
        self._succeed()


class ErrorStep(CommandStep):
    '''Raises an exception to force a rollback. Good for testing.'''
    COMMAND = 'thisCommandDontExist'
    ERROR_CODE = 255



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import codecs
import re
from bag.text import content_of
from grimace import RE  # https://github.com/benlast/grimace/wiki/Documentation


VERSION_NUMBER = str(RE().one_or_more.digits.dot.one_or_more.digits
    .then.any_number_of.any_of('0123456789-.()abcdefghijklmnopqrstuvwxyz'))
VERSION_NUMBER_RE = re.compile(VERSION_NUMBER)
VERSION_VALIDATOR = '^' + VERSION_NUMBER + '$'
VERSION_VALIDATOR_RE = re.compile(VERSION_VALIDATOR)


def error_in_version(val, allow_dev=True):
    '''Validates a version number. Returns None if valid.'''
    if not VERSION_VALIDATOR_RE.match(val):
        return '"{0}" is not a valid version number.'.format(val)
    if not allow_dev:
        for x in ('svn', 'dev', '('):
            if x in val:
                return '"{0}" is a development version number.'.format(val)


SOME_QUOTE = r'["\']'  # single or double quote
QUOTED_VERSION = SOME_QUOTE + '(' + VERSION_NUMBER + ')' + SOME_QUOTE


def version_in_python_source(text, replace=None, keyword='version'):
    '''If ``replace`` is None, returns the version number found in ``text``.
    Else, returns an updated ``text`` (with the version number specified in
    the ``replace`` argument).
    '''
    PYTHON_VERSION_LINE = str(RE().zero_or_more.whitespace
        .between(0, 2).underscore.then(keyword).then.between(0, 2).underscore
        .then.zero_or_more.whitespace.then('=').then.zero_or_more.whitespace
        .then.regex(QUOTED_VERSION).then.zero_or_more.whitespace)
    PYTHON_VERSION_LINE_RE = re.compile(PYTHON_VERSION_LINE)
    match = PYTHON_VERSION_LINE_RE.search(text)
    assert match, 'Could not find version number in Python source code.'
    version = match.groups()[0]
    if replace:
        alist = []
        found = False
        for line in text.splitlines(True):  # True keeps line endings
            match = PYTHON_VERSION_LINE_RE.match(line)
            if not found and match:
                found = True
                line = line.replace(version, replace)
            alist.append(line)
        assert found, 'Could not replace version number in Python source code.'
        return ''.join(alist)
    else:
        return version


def version_in_python_source_file(path, replace=None, keyword='version',
                                  encoding='utf-8'):
    '''If replace is None, returns the version number found in file ``path``.
    Else, replaces the version number with the one specified in ``replace``.
    '''
    ret = version_in_python_source(content_of(path, encoding=encoding),
                                   replace=replace, keyword=keyword)
    if replace:
        with codecs.open(path, 'w', encoding=encoding) as stream:
            stream.write(ret)
        # Test the change we just did by reading the file we just wrote
        assert replace == version_in_python_source_file(
            path, keyword=keyword, encoding=encoding)
    else:
        return ret



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from . import ReleaseStep, CommandStep, StopRelease

__all__ = ('get_current_branch', 'EnsureGitClean', 'EnsureGitBranch',
    'GitCommitVersionNumber', 'GitTag', 'GitPush', 'GitPushTags')


class EnsureGitClean(ReleaseStep):
    '''Ensures we are NOT in a detached head and
    there are no uncommitted changes in tracked files.
    '''
    ERROR_CODE = 52

    def __call__(self):
        head = self._execute_or_complain('git symbolic-ref --quiet HEAD')
        # This returns something like 'refs/heads/master' or nothing.
        # Nothing is bad, indicating a detached head: likely a tag checkout
        if not head:
            raise StopRelease('Wait, are you on a detached head?')
        changes = self._execute_or_complain(
            'git status --short --untracked-files=no')  # sets self.success
        if changes:
            raise StopRelease(
                'There are uncommitted changes in tracked files.')
        self._succeed()


def get_current_branch():
    with open('.git/HEAD') as f:
        return f.read().strip().split('/')[-1]


class EnsureGitBranch(ReleaseStep):
    '''I must be in the branch specified in config.'''
    ERROR_CODE = 51

    def __call__(self):
        required = self.config.get('branch', 'master')
        branch = get_current_branch()
        if branch != required:
            raise StopRelease('You are in branch "{0}", but should be '
                'in branch "{1}" in order to make a release.'.format(
                    branch, required))
        self._succeed()


class GitCommitVersionNumber(ReleaseStep):
    '''Creates a git commit with only one alteration: the new version number.
    Can rollback().
    '''
    COMMAND = 'git commit -a -m "{0}"'
    ERROR_CODE = 53

    def __init__(self, which='the_version', msg='Version {0}',
                 stop_on_failure=True):
        assert which in ('the_version', 'future_version')
        self.which = which
        self.msg = msg.replace('"', '\"')  # escape quotes for shell
        self.stop_on_failure = stop_on_failure

    def __call__(self):
        msg = self.msg.format(getattr(self.releaser, self.which))
        self._execute_or_complain(self.COMMAND.format(msg))  # sets success

    def rollback(self):
        self._execute_or_complain('git reset --hard HEAD^')


class GitTag(ReleaseStep):
    '''Tags the current git commit with the new version number. Can rollback().
    '''
    COMMAND = 'git tag -a v{0} -m "Version {0}"'
    ERROR_CODE = 54
    stop_on_failure = False

    def __call__(self):
        version = self.releaser.the_version
        if version is None:
            self.log.warning('Skipping the GitTag step. It can only run AFTER '
                'some other step sets *the_version* on the releaser.')
            return
        self._execute_or_complain(self.COMMAND.format(version))  # sets success
        if self.success:
            self.releaser.created_tags.append(version)

    def rollback(self):
        self._execute_or_complain('git tag -d "v{0}"'.format(
            self.releaser.the_version))


class GitPush(ReleaseStep):
    COMMAND = 'git push'
    ERROR_CODE = 55
    stop_on_failure = False  # It should be easy to 'git push' afterwards
    no_rollback = 'One should never try to undo a git push. Really.\n' \
        'This release process went far -- the push succeeded -- but\n' \
        'something else went wrong after the push. I figure it will\n' \
        'be easier for you to finish the release process manually,\n' \
        'so I will NOT roll back any of the steps that preceded the push.'

    def __call__(self):
        self._execute_or_complain(self.COMMAND)  # sets success
        if self.success:
            # Erase history so all steps before the push won't be rolled back
            self.log.debug('GitPush successful; erasing rollback history.')
            # self.releaser.rewindable.clear()
            self.releaser.rewindable[:] = []  # Python 2.6 has no clear()
            # self.releaser.non_rewindable.clear()
            self.releaser.non_rewindable[:] = []  # Python 2.6 has no clear()


class GitPushTags(CommandStep):
    '''Pushes local tags to the remote repository. Can rollback().'''
    COMMAND = 'git push --tags'
    ERROR_CODE = 56
    stop_on_failure = False

    def rollback(self):
        for tag in self.releaser.created_tags:
            self._execute_or_complain(
                'git push --delete origin "v{0}"'.format(tag))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)
import os
from setuptools import setup, find_packages
# http://peak.telecommunity.com/DevCenter/setuptools#developer-s-guide


def content_of(*files):
    import codecs
    here = os.path.abspath(os.path.dirname(__file__))
    content = []
    for f in files:
        with codecs.open(os.path.join(here, f), encoding='utf-8') as stream:
            content.append(stream.read())
    return '\n'.join(content)

dependencies = ['nine > 0.3.2', 'bag > 0.3.7', 'docutils',
    'grimace', 'path.py', 'requests']

setup(
    name="releaser",
    description='Automates the process of releasing a new version of '
    'some software.',
    long_description=content_of('README.rst', 'CHANGES.rst'),
    url='https://github.com/nandoflorestan/releaser',
    version='0.1.3',
    license='MIT',
    author='Nando Florestan',
    author_email="nandoflorestan@gmail.com",
    install_requires=dependencies,
    classifiers=[  # http://pypi.python.org/pypi?:action=list_classifiers
        # 'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=['python', 'software', 'release', 'releaser', 'automation',
              'git', 'hg', 'setup.py', 'setuptools', 'distribute',
    ],
    packages=find_packages(),
    include_package_data=True,
    # test_suite='releaser.tests',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

'''Framework for releasing Python software with variable steps in the
release process.
'''

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import subprocess
from sys import platform
from pkg_resources import parse_version
from bag.console import bool_input, screen_header
from bag.log import setup_log
from nine import str, nine
from .regex import error_in_version


class StopRelease(RuntimeError):
    '''Release steps should raise this exception to stop the whole program.'''


@nine
class ReleaseStep(object):
    '''Abstract base class for release steps.'''
    ERROR_CODE = 1
    success = None
    stop_on_failure = True

    def __call__(self):
        '''Override this method to do the main work of the release step.
        On error, StopRelease should be raised.
        '''
        raise StopRelease('Step not implemented.')

    def _succeed(self):
        self.success = True

    def _fail(self, msg):
        self.success = False
        if self.stop_on_failure:
            raise StopRelease(msg)
        else:
            self.log.warning(msg + '\nContinuing anyway.')

    def _execute(self, command, input='', shell=True):
        p = subprocess.Popen(command, shell=shell,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             close_fds=not platform.startswith('win'))
        i, o, e = (p.stdin, p.stdout, p.stderr)
        if input:
            i.write(input)
        i.close()
        return_code = p.wait()
        normal_output = o.read().strip().decode('utf-8')
        error_output = e.read().strip().decode('utf-8')
        o.close()
        e.close()
        if normal_output:
            self.log.debug(normal_output)
        if error_output:
            self.log.error(error_output)
        return return_code, normal_output + error_output

    def _execute_or_complain(self, command, input='', shell=True,
                             msg='Command failed with code {code}: {cmd}'):
        return_code, text = self._execute(command)
        if return_code == 0:
            if hasattr(self, '_validate_command_output'):
                if self._validate_command_output(text):
                    self._succeed()
                else:
                    self._fail(msg.format(code=return_code, cmd=command))
            else:
                self._succeed()
        else:
            self._fail(msg.format(code=return_code, cmd=command))
        return text

    def __str__(self):
        return type(self).__name__  # good default identifier for most steps


class CommandStep(ReleaseStep):
    '''Abstract base class for release steps... that executes a command.'''
    def __call__(self):
        self._execute_or_complain(self.COMMAND)  # sets self.success


class Releaser(object):
    def __init__(self, config, *steps):
        self.created_tags = []
        self.config = config
        # Configure logging
        path = config.get('log_file', 'release.log.utf-8.tmp')
        screen_level = config.get('verbosity', 'info')
        self.log = setup_log(path=path, rotating=False, file_mode='w',
                             disk_level='debug', screen_level=screen_level)
        # Convert steps provided by the user into real instances
        self.instances = []
        for step in steps:
            if isinstance(step, type):
                step = step()  # Instantiate the ReleaseStep subclass
            step.config = config
            step.releaser = self
            step.log = self.log
            self.instances.append(step)
        # Now that all steps are correctly instantiated,
        # it is safe to start running them by calling release().

    def release(self):
        self.rewindable = []
        self.non_rewindable = []
        for step in self.instances:
            self.log.info(screen_header(step))
            try:
                step()
            except StopRelease as e:
                self.log.critical('Release process stopped at step {0}:\n{1}'
                    .format(step, e))
                self.rewind()
                from sys import exit
                exit(step.ERROR_CODE)
            except Exception:
                self.log.debug('Noooo!!!', exc_info=True)
                self.rewind()
                raise
            else:
                if step.success and hasattr(step, 'no_rollback'):
                    self.non_rewindable.append(step.no_rollback)
                elif step.success and hasattr(step, 'rollback'):
                    self.rewindable.append(step)
        self.log.info('Successfully released version {0}. '
            'Sorry for the convenience, mcdonc!'.format(self.the_version))

    def rewind(self):
        say = self.log.critical
        say('\n' + screen_header('ROLLBACK', decor='*'))
        # Some steps offer a warning because they cannot be rolled back.
        say('\n'.join(self.non_rewindable))              # Display them.
        if not self.rewindable:
            say('No steps to roll back, but the release process FAILED.')
            return
        steps = list(reversed(self.rewindable))
        print('I am about to roll back the following steps:\n{0}'.format(
            ', '.join([str(step) for step in steps])))
        if not bool_input('Continue?', default=True):
            return
        for step in steps:
            self.log.critical(screen_header('ROLLBACK {0}'.format(step)))
            try:
                step.rollback()
            except Exception as e:
                self.log.error('Could not roll back step {0}:\n{1}'
                    .format(step, str(e)))

    _old_version = None    # 0.1.2dev (exists when the program starts)
    _the_version = None    # 0.1.2    (the version being released)

    @property
    def old_version(self):
        return self._old_version

    @old_version.setter
    def old_version(self, val):
        self._old_version = val
        self.log.debug('Old version: {0}'.format(val))

    @property
    def the_version(self):
        return self._the_version

    @the_version.setter
    def the_version(self, val):
        val = val.strip()
        e = error_in_version(val, allow_dev=False)
        if e:
            raise StopRelease(e)
        if parse_version(self.old_version) >= parse_version(val):
            raise StopRelease(
                'No, the version number must be higher than the current one!')
        self._the_version = val
        self.log.debug('Version being released: {0}'.format(val))

    @property
    def future_version(self):  # 0.1.3dev (development version after release)
        parts = self.the_version.split('.')
        parts[-1] = str(int(parts[-1]) + 1)
        return '.'.join(parts) + 'dev'
